# -*- coding: utf-8 -*-

"""A Tk scrolled frame widget compatible with tkinter.ttk

Based on a version:
# Version: 0.22
# Author: Miguel Martinez Lopez
# Uncomment the next line to see my email
# print("Author's email: ",
# "61706c69636163696f6e616d656469646140676d61696c2e636f6d".decode("hex"))

with minor changes.
"""

import seamm_widgets as sw
import tkinter as tk
from tkinter import ttk


class ScrolledLabelFrame(ttk.LabelFrame):
    def __init__(
        self,
        master,
        text=None,
        labelanchor=None,
        labelwidget=None,
        underline=None,
        width=None,
        anchor=tk.N,
        height=None,
        mousewheel_speed=2,
        scroll_horizontally=True,
        xscrollbar=None,
        scroll_vertically=True,
        yscrollbar=None,
        background=None,
        inner_frame=ttk.Frame,
        **kwargs
    ):
        """ """
        class_ = kwargs.pop("class_", "MScrolledLabelFrame")
        super().__init__(master, class_=class_)

        if text is not None:
            self.configure(text=text)
        if labelanchor is not None:
            self.configure(labelanchor=labelanchor)
        if labelwidget is not None:
            self.configure(labelwidget=labelwidget)
        if underline is not None:
            self.configure(underline=underline)

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self._width = width
        self._height = height

        self.canvas = tk.Canvas(
            self,
            background=background,
            highlightthickness=0,
            width=width,
            height=height,
            takefocus=True,
        )
        self.canvas.grid(row=0, column=0, sticky=tk.NSEW)

        if scroll_vertically:
            if yscrollbar is not None:
                self.yscrollbar = yscrollbar
            else:
                self.yscrollbar = ttk.Scrollbar(
                    self, orient=tk.VERTICAL, takefocus=False
                )
                self.yscrollbar.grid(row=0, column=1, sticky=tk.NS)

            self.canvas.configure(yscrollcommand=self.yscrollbar.set)
            self.yscrollbar["command"] = self.canvas.yview
        else:
            self.yscrollbar = None

        if scroll_horizontally:
            if xscrollbar is not None:
                self.xscrollbar = xscrollbar
            else:
                self.xscrollbar = ttk.Scrollbar(
                    self, orient=tk.HORIZONTAL, takefocus=False
                )
                self.xscrollbar.grid(row=1, column=0, sticky=tk.EW)

            self.canvas.configure(xscrollcommand=self.xscrollbar.set)
            self.xscrollbar["command"] = self.canvas.xview
        else:
            self.xscrollbar = None

        self.rowconfigure(0, weight=1)
        self.columnconfigure(0, weight=1)

        self.innerframe = inner_frame(self.canvas, **kwargs)
        self.innerframe.pack(anchor=anchor)

        self.canvas.create_window(
            0, 0, window=self.innerframe, anchor="nw", tags="inner_frame"
        )

        self.canvas.bind("<Configure>", self._on_canvas_configure)

        sw.MousewheelSupport(self).add_support_to(
            self.canvas, xscrollbar=self.xscrollbar, yscrollbar=self.yscrollbar
        )

    @property
    def width(self):
        return self.canvas.winfo_width()

    @width.setter
    def width(self, width):
        self.canvas.configure(width=width)

    @property
    def height(self):
        return self.canvas.winfo_height()

    @height.setter
    def height(self, height):
        self.canvas.configure(height=height)

    def interior(self):
        """The frame that contains user widgets"""
        return self.innerframe

    def set_size(self, width, height):
        self.canvas.configure(width=width, height=height)

    def _on_canvas_configure(self, event):
        width = max(self.innerframe.winfo_reqwidth(), event.width)
        height = max(self.innerframe.winfo_reqheight(), event.height)

        self.canvas.configure(scrollregion="0 0 %s %s" % (width, height))
        self.canvas.itemconfigure("inner_frame", width=width, height=height)

    def update_viewport(self):
        self.update()

        window_width = self.innerframe.winfo_reqwidth()
        window_height = self.innerframe.winfo_reqheight()

        if self._width is None:
            canvas_width = window_width
        else:
            canvas_width = min(self._width, window_width)

        if self._height is None:
            canvas_height = window_height
        else:
            canvas_height = min(self._height, window_height)

        self.canvas.configure(
            scrollregion="0 0 %s %s" % (window_width, window_height),
            width=canvas_width,
            height=canvas_height,
        )
        self.canvas.itemconfigure(
            "inner_frame", width=window_width, height=window_height
        )



# -*- coding: utf-8 -*-

"""A Tk widget for a tree of checkboxes."""

import logging
from pathlib import Path
import pkg_resources
import tkinter as tk
from tkinter import ttk

from PIL import ImageTk, Image

import seamm_widgets as sw

logger = logging.getLogger(__name__)

options = {
    "treeview": {
        "columns": "columns",
        "cursor": "cursor",
        "displaycolumns": "displaycolumns",
        "height": "height",
        "padding": "padding",
        "show": "show",
        "style": "style",
        "takefocus": "takefocus",
    },
}


class CheckTree(sw.LabeledWidget):
    """Class to provide a tree of checkboxes."""

    def __init__(self, parent, *args, columns=[], **kwargs):
        class_ = kwargs.pop("class_", "MCheckTree")
        super().__init__(parent, class_=class_)

        self.frame = self.interior
        self._columns = columns

        self.tree = ttk.Treeview(
            self.frame,
            selectmode="none",
            columns=self._columns,
        )
        for column in self._columns:
            self.tree.heading(column, text=column)

        self.tree.tag_bind("_node_", "<Button-1>", self._click_cb)

        self.x_scrollbar = ttk.Scrollbar(
            self.frame, orient=tk.HORIZONTAL, command=self.tree.xview
        )
        self.tree.config(xscrollcommand=self.x_scrollbar.set)
        self.y_scrollbar = ttk.Scrollbar(
            self.frame, orient=tk.VERTICAL, command=self.tree.yview
        )
        self.tree.config(yscrollcommand=self.y_scrollbar.set)

        self.tree.grid(row=0, column=0, sticky=tk.NSEW)
        self.y_scrollbar.grid(row=0, column=1, sticky=tk.NS)
        self.x_scrollbar.grid(row=1, column=0, sticky=tk.EW)
        self.frame.rowconfigure(0, weight=1)
        self.frame.columnconfigure(0, weight=1)

        path = Path(pkg_resources.resource_filename(__name__, "data/"))

        self.checked_image = ImageTk.PhotoImage(Image.open(path / "checked.png"))
        self.unchecked_image = ImageTk.PhotoImage(Image.open(path / "unchecked.png"))
        self.mixed_image = ImageTk.PhotoImage(Image.open(path / "mixed.png"))

        self.config(**kwargs)

    def bbox(self, item, column=None):
        """Return the bonding box for the item.

        For the item with iid item, if the item is currently visible, this method
        returns a tuple (x, y, w, h), where (x, y) are the coordinates of the upper left
        corner of that item relative to the widget, and w and h are the width and height
        of the item in pixels. If the item is not visible, the method returns an empty
        string.

        If the optional column argument is omitted, you get the bounding box of the
        entire row. To get the bounding box of one specific column of the item's row,
        use column=C where C is either the integer index of the column or its column
        identifier.
        """
        return self.tree.bbox(item, column=column)

    def column(self, cid, option=None, **kw):
        """
        This method configures the appearance of the logical column specified by cid,
        which may be either a column index or a column identifier. To configure the icon
        column, use a cid value of '#0'.

        Each column in a Treeview widget has its own set of options from this table:

            anchor
                The anchor that specifies where to position the content of the
                column. The default value is 'w'.

            id
                The column name. This option is read-only and set when the constructor
                is called.

            minwidth
                Minimum width of the column in pixels; the default value is 20.

            stretch
                If this option is True, the column's width will be adjusted when the
                widget is resized. The default setting is 1.

            width
                Initial width of the column in pixels; the default is 200.

        If no option value or any other keyword argument is supplied, the method returns
        a dictionary of the column options for the specified column.

        To interrogate the current value of an option named X , use an argument
        option=X.

        To set one or more column options, you may pass keyword arguments using the
        option names shown above, e.g., anchor=tk.CENTER to center the column contents.
        """
        return self.tree.column(cid, option=option, **kw)

    def config(self, **kwargs):
        """Set the configuration of the megawidget"""

        # our options that we deal with
        entry = options["treeview"]

        # cannot modify kwargs while iterating over it...
        keys = [*kwargs.keys()]
        for k in keys:
            if k in entry:
                v = kwargs.pop(k)
                self.tree.config(**{entry[k]: v})

        # having removed our options, pass rest to parent
        super().config(**kwargs)

    def configure(self, **kwargs):
        """Alias for config()"""
        return self.config(**kwargs)

    def _check_parents(self, iid):
        """Correct the state of all parents up the tree to make their state correct.

        Parameters
        ----------
        iid : str
            The item to start with.
        """
        parent = self.tree.parent(iid)
        logger.debug(f"{iid=} {parent=}")
        if parent == "":
            self.state(iid)
        else:
            self._check_parents(parent)

    def _click_cb(self, event):
        iid = self.tree.identify_row(event.y)

        element = self.tree.identify_element(event.x, event.y)

        logger.info(f"_click_cb, {iid=} {element=}")

        if element in ("image", "text"):
            self.invoke(iid)

    def delete(self, *items):
        """The arguments are iid values. All the items in the widget that have matching
        iid values are destroyed, along with all their descendants.
        """
        return self.tree.delete(*items)

    def _deselect(self, iid, recursive=True, check_parents=True):
        """Change the state of the item 'iid' to unchecked.

        Parameters
        ----------
        iid : str
            The unique id of the item in the tree.
        recursive : bool
            Whether to select all children
        check_parents: bool
            Whether to set the state of parents

        Returns
        -------
        any
            The return from any command invoked by changing the state.
        """
        tags = [*self.tree.item(iid, "tags")]
        if "checked" in tags:
            tags.remove("checked")
        self.tree.item(iid, image=self.unchecked_image, tags=tags)

        # If this is a branch node, i.e. it has children, change them too
        if recursive:
            children = self.tree.get_children(iid)
            for child in children:
                self._deselect(child, recursive=True)

        if check_parents:
            self._check_parents(iid)

    def detach(self, *items):
        """The arguments are iid values. All the items in the widget that have matching
        iid values are removed from the visible widget, along with all their
        descendants.

        The items are not destroyed. You may reattach them to the visible tree using the
        .move() method described below.
        """
        return self.tree.detach(*items)

    def exists(self, iid):
        """Returns True if there exists an item in the widget with the given iid, or
        False otherwise. If an item is not currently visible because it was removed with
        the .detach() method, it is still considered to exist for the purposes of the
        .exists() method.
        """
        return self.tree.exists(iid)

    def from_dict(self, parent, metadata):
        for key, value in metadata.items():
            # Get any data in the columns
            column_data = []
            for column in self._columns:
                tmp = ""
                for column_name in value.keys():
                    if column_name.lower() == column.lower():
                        tmp = value[column_name]
                        break
                column_data.append(tmp)

            tags = ["_node_", key]
            checked = value.get("checked", False)
            if checked:
                tags.append("checked")

            self.tree.insert(
                parent,
                "end",
                iid=key,
                text=key,
                image=self.checked_image if checked else self.unchecked_image,
                tag=tags,
                open="open" in value and value["open"],
                values=column_data,
            )
            # Recurse if a branch node
            if "items" in value:
                self.from_dict(key, value["items"])

        # Set the state of intermediate nodes
        children = self.tree.get_children("")
        for child in children:
            self.state(child)

    def focus(self, iid=None):
        """If you don't provide an argument to this method, you get back either the iid
        of the item that currently has focus, or '' if no item has focus.

        You can give focus to an item by passing its iid as the argument to this method.
        """
        return self.tree.focus(iid)

    def get(self, parent="", as_dict=False):
        """Get the selected items.

        Parameters
        ----------
        parent : str = ""
            Get selected children of the parent. Defaults to the entire tree.
        as_dict : bool = False
            Return a dictionary keyed by the branching nodes, rather than a flat list.

        Returns
        -------
        [str] or dict(str: str|dict(str: str|dict(...)
        """
        if as_dict:
            result = {"selected": [], "children": {}}
        else:
            result = []
        for child in self.get_children(parent):
            if len(self.get_children(child)) > 0:
                if as_dict:
                    result["children"][child] = self.get(parent=child, as_dict=as_dict)
                else:
                    result.extend(self.get(parent=child))
            else:
                tags = self.item(child, "tags")
                if "checked" in tags:
                    if as_dict:
                        result["selected"].append(child)
                    else:
                        result.append(child)
        return result

    def get_children(self, item=None):
        """Returns a tuple of the iid values of the children of the item specified by
        the item argument. If the argument is omitted, you get a tuple containing the
        iid values of the top-level items.
        """
        return self.tree.get_children(item)

    def get_leaves(self, item=None):
        """Returns the iid values of all leaves of the tree belove of the item specified
        by the item argument. If the argument is omitted, it gives the leaves for the
        entire tree.
        """
        leaves = []
        for child in self.tree.get_children(item):
            children = self.tree.get_children(child)
            if len(children) > 0:
                leaves.extend(self.get_leaves(child))
            else:
                leaves.append(child)
        return leaves

    def heading(self, cid, option=None, **kw):
        """Use this method to configure the column heading that appears at the top of
        the widget for the column specified by cid, which may be either a column index
        or a column identifier. Use a cid argument value of '#0' to configure the
        heading over the icon column.

        Each heading has its own set of options with these names and values:

        anchor
            An anchor that specifies how the heading is aligned within the column; see
            Section 5.5, “Anchors”. The default value is tk.W.
        command
            A procedure to be called when the user clicks on this column heading.
        image
            To present a graphic in the column heading (either with or instead of a text
            heading), set this option to an image, as specified in Section 5.9,
            “Images”.
        text
            The text that you want to appear in the column heading.

        If you supply no keyword arguments, the method will return a dictionary showing
        the current settings of the column heading options.

        To interrogate the current value of some heading option X, use an argument of
        the form option=X; the method will return the current value of that option.

        You can set one or more heading options by supplying them as keyword arguments
        such as “anchor=tk.CENTER”.
        """
        return self.tree.heading(cid, option=option, **kw)

    def identify_column(self, x):
        """Given an x coordinate, this method returns a string of the form '#n' that
        identifies the column that contains that x coordinate.

        Assuming that the icon column is displayed, the value of n is 0 for the icon
        column; 1 for the second physical column; 2 for the third physical column; and
        so on. Recall that the physical column number may be different from the logical
        column number in cases where you have rearranged them using the displaycolumns
        argument to the Treeview constructor.

        If the icon column is not displayed, the value of n is 1 for the first physical
        column, 2 for the second, and so on.
        """
        return self.tree.identify_column(x)

    def identify_element(self, x, y):
        """Returns the name of the element at location (x, y) relative to the widget, or
        '' if no element appears at that position. Element names are discussed in
        Section 50, “The ttk element layer”.
        """
        return self.tree.identify_element(x, y)

    def identify_region(self, x, y):
        """Given the coordinates of a point relative to the widget, this method returns
        a string indicating what part of the widget contains that point. Return values
        may include:

        'nothing'
            The point is not within a functional part of the widget.
        'heading'
            The point is within one of the column headings.
        'separator'
             The point is located within the column headings row, but on the separator
             between columns. Use the .identify_column() method to determine which
             column is located just to the left of this separator.
        'tree'
             The point is located within the icon column.
        'cell'
             The point is located within an item row but not within the icon column.
        """
        return self.tree.identify_region(x, y)

    def identify_row(self, y):
        """If y-coordinate y is within one of the items, this method returns the iid of
        that item. If that vertical coordinate is not within an item, this method
        returns an empty string.
        """
        return self.tree.identify_row(y)

    def index(self, iid):
        """This method returns the index of the item with the specified iid relative to
        its parent, counting from zero.
        """
        return self.tree.index(iid)

    def insert(
        self,
        parent,
        index,
        iid=None,
        open=False,
        selected=False,
        state=True,
        tags=[],
        text="",
        values=[],
    ):
        """This method adds a new item to the tree, and returns the item's iid
        value.

        Parameters
        ----------
        parent : str
            To insert a new top-level item, make this argument an empty string. To
            insert a new item as a child of an existing item, make this argument the
            parent item's iid.
        index : int or 'end'
            This argument specifies the position among this parent's children where you
            want the new item to be added. For example, to insert the item as the new
            first child, use a value of zero; to insert it after the parent's first
            child, use a value of 1; and so on. To add the new item as the last child of
            the parent, make this argument's value 'end'.
        iid : str = None
            You may supply an iid for the item as a string value. If you don't supply an
            iid, one will be generated automatically and returned by the method.
        open : bool
            This option specifies whether this item will be open initially. If you
            supply open=False, this item will be closed. If you supply open=True, the
            item's children will be visible whenever the item itself is visible. The
            default value is False.
        selected : bool = False
            Whether the item is initially selected.
        state : bool = True
            Whether to correct the state of the tree, default=True
        tags : [str] = []
            You may supply one or more tag strings to be associated with this item. The
            value may be either a single string or a sequence of strings.
        text : str = ""
            You may supply text to be displayed within the icon column of this item. If
            given, this text will appear just to the right of the icon, and also to the
            right of the image if provided.
        values : [str] = []
            This argument supplies the data items to be displayed in each column of the
            item. The values are supplied in logical column order. If too few values are
            supplied, the remaining columns will be blank in this item; if too many
            values are supplied, the extras will be discarded.
        """
        tmp_tags = [*tags]
        if "_node_" not in tmp_tags:
            tmp_tags.append("_node_")

        iid = self.tree.insert(
            parent,
            index,
            iid=iid,
            open=open,
            tags=tmp_tags,
            text=text,
            values=values,
            image=self.checked_image if selected else self.unchecked_image,
        )

        if selected:
            self.selection_add(iid)
        if state:
            self.state()
        return iid

    def invoke(self, iid, recursive=True):
        """Change the state of the item 'iid' just as if clicked.

        Parameters
        ----------
        iid : str
            The unique id of the item in the tree.
        recursive : bool
            Change all children, grandchildren, etc. if True

        Returns
        -------
        any
            The return from any command invoked by changing the state.
        """
        tags = [*self.tree.item(iid, "tags")]
        if "checked" in tags:
            self._deselect(iid, recursive=recursive)
        else:
            self._select(iid, recursive=recursive)

    def item(self, iid, option=None, **kw):
        """Use this method to set or retrieve the options within the item specified by
        iid. Refer to the .insert() method above for the names of the item options.

        With no arguments, it returns a dictionary whose keys are the option names and
        the corresponding values are the settings of those options. To retrieve the
        value of a given option, pass the option's name as its second argument. To set
        one or more options, pass them as keyword arguments to the method.
        """
        return self.tree.item(iid, option=option, **kw)

    def move(self, iid, parent, index):
        """Move the item specified by iid to the values under the item specified by
        parent at position index. The parent and index arguments work the same as those
        arguments to the .index() method.
        """
        return self.tree.move(iid, parent, index)

    def next(self, iid):
        """If the item specified by iid is not the last child of its parent, this method
        returns the iid of the following child; if it is the last child of its parent,
        this method returns an empty string. If the specified item is a top-level item,
        the method returns the iid of the next top-level item, or an empty string if the
        specified item is the last top-level item.
        """
        return self.tree.next(iid)

    def parent(self, iid):
        """If the item specified by iid is a top-level item, this method returns an
        empty string; otherwise it returns the iid of that item's parent.
        """
        return self.tree.parent(iid)

    def prev(self, iid):
        """If the item specified by iid is not the first child of its parent, this
        method returns the iid of the previous child; otherwise it returns an empty
        string. If the specified item is a top-level item, this method returns the iid
        of the previous top-level item, or an empty string if it is the first top-level
        item.
        """
        return self.tree.prev(iid)

    def see(self, iid):
        """This method ensures that the item specified by iid is visible. Any of its
        ancestors that are closed are opened. The widget is scrolled, if necessary, so
        that the item appears.
        """
        return self.tree.see(iid)

    def selection_add(self, items):
        """In addition to any items already selected, add the specified items. The
        argument may be either a single iid or a sequence of iids.
        """
        if isinstance(items, str):
            self._select(items)
        else:
            for item in items:
                self._select(item)

    def selection_clear(self, items=None):
        """Clear all selections from the subtree(s)"""
        if items is None:
            self.selection_remove("")
        elif isinstance(items, str):
            self.selection_remove(self.get(items))
        else:
            for item in items:
                self.selection_remove(self.get(item))

    def selection_remove(self, items):
        """Unselect any items specified by the argument, which may be a single iid or a
        sequence of iids.
        """
        if isinstance(items, str):
            self._deselect(items)
        else:
            for item in items:
                self._deselect(item)

    def selection_set(self, items):
        """Only the specified items will be selected; if any other items were selected
        before, they will become unselected.
        """
        self.selection_remove(self.get_leaves())
        self.selection_add(items)

    def selection_toggle(self, items):
        """The argument may be a single iid or a sequence of iids. For each item
        specified by the argument, if it was selected, unselect it; if it was
        unselected, select it.
        """
        if isinstance(items, str):
            self.invoke(items)
        else:
            for item in items:
                self.invoke(item)

    def _select(self, iid, recursive=True, check_parents=True):
        """Change the state of the item 'iid' to checked.

        Parameters
        ----------
        iid : str
            The unique id of the item in the tree.
        recursive : bool
            Whether to select all children
        check_parents: bool
            Whether to set the state of parents

        Returns
        -------
        any
            The return from any command invoked by changing the state.
        """
        tags = [*self.tree.item(iid, "tags")]
        tags.append("checked")
        self.tree.item(iid, image=self.checked_image, tags=tags)

        # If this is a branch node, i.e. it has children, change them too
        if recursive:
            children = self.tree.get_children(iid)
            for child in children:
                self._select(child, recursive=True, check_parents=False)

        if check_parents:
            self._check_parents(iid)

    def set(self, iid, column=None, value=None):
        """Use this method to retrieve or set the column values of the item specified by
        iid. With one argument, the method returns a dictionary: the keys are the column
        identifiers, and each related value is the text in the corresponding column.

        With two arguments, the method returns the data value from the column of the
        selected item whose column identifier is the column argument. With three
        arguments, the item's value for the specified column is set to the third
        argument.
        """
        return self.tree.set(iid, column=column, value=value)

    def set_children(self, item, *newChildren):
        """Use this method to change the set of children of the item whose iid is
        item. The newChildren argument is a sequence of iid strings. Any current
        children of item that are not in newChildren are removed.
        """
        return self.tree.set_children(item, *newChildren)

    def state(self, iid="", recursive=True):
        """Determine the state of this item, and optionally its subtree.

        Parameters
        ----------
        iid : str
            The item to check, or entire tree if ""
        recursive : bool
            Whether to check the subtree

        Returns
        -------
        str
            "checked", "unchecked", or "mixed"
        """
        state = None
        logger.debug(f"\n state for {iid}")
        if recursive:
            logger.debug(f"{recursive=}")
            if iid is None:
                children = self.get_children("")
            else:
                children = self.get_children(iid)
            for child in children:
                tmp = self.state(child, recursive=True)
                if state is None:
                    state = tmp
                    logger.debug(f"    {child=} {state=}")
                elif state != tmp:
                    logger.debug(f"    {child=} returning 'mixed'")
                    state = "mixed"
            if state is not None:
                tags = [*self.tree.item(iid, "tags")]
                logger.debug(f"{state=} {tags=}")
                if state == "checked":
                    if "checked" not in tags:
                        tags.append("checked")
                    if "mixed" in tags:
                        tags.remove("mixed")
                    self.tree.item(iid, image=self.checked_image, tags=tags)
                elif state == "unchecked":
                    if "checked" in tags:
                        tags.remove("checked")
                    if "mixed" in tags:
                        tags.remove("mixed")
                    self.tree.item(iid, image=self.unchecked_image, tags=tags)
                else:
                    if "checked" in tags:
                        tags.remove("checked")
                    if "mixed" not in tags:
                        tags.append("mixed")
                    self.tree.item(iid, image=self.mixed_image, tags=tags)

                logger.debug(f"returning {state=} from branch node {iid=}")
                return state

        tags = [*self.tree.item(iid, "tags")]
        logger.debug(f"    {tags=}")
        if "checked" in tags:
            state = "checked"
        elif "mixed" in tags:
            state = "mixed"
        else:
            state = "unchecked"

        logger.debug(f"returning {state=} from leaf node {iid=}")
        return state

    def tag_bind(self, tagName, sequence=None, callback=None):
        """This method binds the event handler specified by the callback argument to all
        items that have tag tagName. The sequence and callback arguments work the same
        as the sequence and func arguments of the .bind() method described in Section
        26, “Universal widget methods”.
        """
        return self.tree.tag_bind(tagName, sequence=sequence, callback=callback)

    def tag_configure(self, tagName, option=None, **kw):
        """This method can either interrogate or set options that affect the appearance
        of all the items that have tag tagName. Tag options include:

        'background'
            The background color.
        'font'
            The text font.
        'foreground'
            The foreground color.
        'image'
            An image to be displayed in items with the given tag.

        When called with one argument, it returns a dictionary of the current tag
        options. To return the value of a specific option X, use X as the second
        argument.

        To set one or more options, use keyword arguments such as foreground='red'.
        """
        return self.tree.tag_configure(tagName, option=option, **kw)

    def tag_has(self, tagName, iid=None):
        """Called with one argument, this method returns a list of the iid values for
        all items that carry tag tagName. If you provide an iid as the second argument,
        the method returns True if the item with that iid has tag tagName, False
        otherwise.
        """
        return self.tree.tag_has(tagName, iid=iid)


if __name__ == "__main__":  # pragma: no cover
    import pprint  # noqa: F401

    # logging.basicConfig(level="INFO")

    def load_dict(tree, parent, metadata):
        """Custom code to load the dictionary in the test.
        NB. the 'state=False' argument prevents insert from working out the
        state of the parent nodes every time an item is inserted. This is done
        at the end with 'tree.state()'
        """
        for key, value in metadata.items():
            # Get any data in the columns
            checked = value.get("checked", False)

            tree.insert(
                parent,
                "end",
                iid=key,
                text=key,
                selected=checked,
                state=False,
                open="open" in value and value["open"],
                values=[value["description"] if "description" in value else ""],
            )
            # Recurse if a branch node
            if "items" in value:
                load_dict(tree, key, value["items"])

        # Set the state of intermediate nodes
        tree.state()

    root = tk.Tk()
    root.title("Search Criteria")

    tree = CheckTree(root, labeltext="Features", labelpos="n", columns=["Description"])
    tree.pack(expand=True, fill="both")

    metadata = {
        "1-D": {
            "open": True,
            "items": {
                "D1": {"description": "This describes D1"},
                "D2": {"checked": True, "description": "This describes D2"},
                "Sub1": {
                    "items": {
                        "S1": {"description": "This describes S1"},
                        "S2": {"item": "alias for S2", "description": "Ain't none!"},
                    },
                },
                "D3": {"item": "D3"},
            },
        },
    }

    if False:
        metadata["2-D"] = {"items": {}}
        tmp = metadata["2-D"]["items"]
        for i in range(20):
            tmp[f"E{i}"] = {"description": f"And a super E{i} thingamajig"}
        load_dict(tree, "", metadata)
    else:
        load_dict(tree, "", metadata)
        twoD = tree.insert("", "end", text="2-D", values=["Two dimensional features"])
        print(f"{twoD=}")
        for i in range(20):
            tree.insert(
                twoD,
                "end",
                text=f"2-D Descriptor {i}",
                values=[f"Fancy 2-D descriptor #{i}"],
            )

    # enter the event loop
    root.mainloop()



# -*- coding: utf-8 -*-

"""A Tk widget for gathering criteria for e.g. a search.

This widget provides a table of criteria, plus a '+' button to add another
row to the table. Each criterion is implemented as a "Criterion" widget, which
has the following subwidgets:

    inclusion
        A combobox giving an introductory condition, such as 'with' or 'without',
        or 'contains' or 'does not contain'.
    field
        A combobox giving the field to search, e.g. 'keyword' or 'author'
    operator
        A combobox giving the search operator, e.g. '<', '>' or '='
    value
        An entry for the value of the, e.g. 'molecule' or 'Doe, John'
    value2
        A second optional entry for operators such as 'between'

Each of these subwidgets is itself a labeled widget, giving more flexibility
in composing a reasonable natural language statement, such as:

    entries that contain a keyword like molecule
    entries that do not contain a author like Doe, John
"""
import logging
import tkinter as tk
from tkinter import ttk

import seamm_widgets as sw

logger = logging.getLogger(__name__)


class Criterion(sw.LabeledWidget):
    """Class to provide a line in a search criteria widget."""

    def __init__(self, parent, *args, **kwargs):
        class_ = kwargs.pop("class_", "MCriterion")
        super().__init__(parent, class_=class_)

        frame = self.frame = self.interior

        self.command = kwargs.pop("command", None)

        # Inclusion combobox
        inclusiontext = kwargs.pop("inclusiontext", "")
        inclusionvalues = kwargs.pop("inclusionvalues", ("and", "or"))
        inclusionheight = kwargs.pop("inclusionheight", 7)
        inclusionwidth = kwargs.pop("inclusionwidth", 3)
        inclusionstate = kwargs.pop("inclusionstate", "readonly")

        self.inclusion = sw.LabeledCombobox(
            frame,
            labeltext=inclusiontext,
            values=inclusionvalues,
            height=inclusionheight,
            width=inclusionwidth,
            state=inclusionstate,
        )
        self.inclusion.grid(row=0, column=0, sticky=tk.EW)
        if len(inclusionvalues) > 0:
            self.inclusion.set(inclusionvalues[0])

        cb = self.inclusion.combobox
        cb.bind(
            "<<ComboboxSelected>>",
            lambda event, what="inclusion": self.callback(event, what),
        )
        if inclusionstate == "normal":
            cb.bind(
                "<Return>", lambda event, what="inclusion": self.callback(event, what)
            )
            cb.bind(
                "<FocusOut>", lambda event, what="inclusion": self.callback(event, what)
            )

        # Field combobox
        fieldtext = kwargs.pop("fieldtext", "")
        fieldvalues = kwargs.pop("fieldvalues", tuple())
        fieldheight = kwargs.pop("fieldheight", 7)
        fieldwidth = kwargs.pop("fieldwidth", 10)
        fieldstate = kwargs.pop("fieldstate", "readonly")

        self.field = sw.LabeledCombobox(
            frame,
            labeltext=fieldtext,
            values=fieldvalues,
            height=fieldheight,
            width=fieldwidth,
            state=fieldstate,
        )
        self.field.grid(row=0, column=1, sticky=tk.EW)
        if len(fieldvalues) > 0:
            self.field.set(fieldvalues[0])

        cb = self.field.combobox
        cb.bind(
            "<<ComboboxSelected>>",
            lambda event, what="field": self.callback(event, what),
        )
        if fieldstate == "normal":
            cb.bind("<Return>", lambda event, what="field": self.callback(event, what))
            cb.bind(
                "<FocusOut>", lambda event, what="field": self.callback(event, what)
            )

        # Operator combobox
        operatortext = kwargs.pop("operatortext", "")
        operatorvalues = kwargs.pop("operatorvalues", tuple())
        operatorheight = kwargs.pop("operatorheight", 7)
        operatorwidth = kwargs.pop("operatorwidth", 10)
        operatorstate = kwargs.pop("operatorstate", "readonly")

        self.operator = sw.LabeledCombobox(
            frame,
            labeltext=operatortext,
            values=operatorvalues,
            height=operatorheight,
            width=operatorwidth,
            state=operatorstate,
        )
        self.operator.grid(row=0, column=2, sticky=tk.EW)
        if len(operatorvalues) > 0:
            self.operator.set(operatorvalues[0])

        cb = self.operator.combobox
        cb.bind(
            "<<ComboboxSelected>>",
            lambda event, what="operator": self.callback(event, what),
        )
        if operatorstate == "normal":
            cb.bind(
                "<Return>", lambda event, what="operator": self.callback(event, what)
            )
            cb.bind(
                "<FocusOut>", lambda event, what="operator": self.callback(event, what)
            )

        # Value entry
        valuetext = kwargs.pop("valuetext", "")
        valuewidth = kwargs.pop("valuewidth", 10)

        self.value = sw.LabeledEntry(frame, labeltext=valuetext, width=valuewidth)
        self.value.grid(row=0, column=3, sticky=tk.EW)

        # Value2 entry
        value2text = kwargs.pop("value2text", " and ")
        value2width = kwargs.pop("value2width", 10)

        self.value2 = sw.LabeledEntry(frame, labeltext=value2text, width=value2width)
        self.value2.grid(row=0, column=4, sticky=tk.EW)

        # interior frame
        self.interior = ttk.Frame(frame)
        self.interior.grid(row=0, column=5, sticky=tk.NSEW)
        frame.columnconfigure(5, weight=1)

        two_values = kwargs.pop("two_values", False)

        self.config(**kwargs)

        self.two_values = two_values

    @property
    def two_values(self):
        """Whether to display two values for e.g. 'between'."""
        return self._two_values

    @two_values.setter
    def two_values(self, value):
        if not isinstance(value, bool):
            try:
                iter(value)
            except TypeError:
                raise ValueError(
                    "'two_values' must be a boolean or iterable of opertors that need "
                    "two values."
                )
        self._two_values = value
        self.callback("internal", "self.two_values")

    def callback(self, event, what):
        """Call the command for changes in the widget value.
        Also check the operator value and reconfigure the widget for one or two
        values if requested."""

        if what in ("operator", "self.two_values", "set"):
            if isinstance(self._two_values, bool):
                show = self._two_values
            else:
                operator = self.operator.get()
                show = operator in self._two_values

            if show:
                self.show("value2")
                # if self.value2 not in self.frame.grid_slaves():
                #     self.value2.grid(row=0, column=4, sticky=tk.EW)
            else:
                self.hide("value2")
                # if self.value2 in self.frame.grid_slaves():
                #     self.value2.grid_forget()
        elif what == "inclusion":
            include = self.inclusion.get()
            if include in ("and", "or", "(", ")"):
                self.hide("all")
                self.show("inclusion")
            else:
                self.show("all")
                if isinstance(self._two_values, bool):
                    show = self._two_values
                else:
                    operator = self.operator.get()
                    show = operator in self._two_values
                if not show:
                    self.hide("value2")

        if self.command is not None:
            self.command(self, event, what)

    def clear(self):
        """Remove all rows from the table."""
        for remove_button, criterion in self._rows.items():
            remove_button.destroy()
            criterion.destroy()

        self.callback(None, "internal", "clear")

        self._rows = []
        self.layout()

    def get(self):
        """Return the values of the widgets as a tuple."""
        return (
            self.inclusion.get(),
            self.field.get(),
            self.operator.get(),
            self.value.get(),
            self.value2.get(),
        )

    def hide(self, *args):
        """Hide the specified subwidgets.
        'all' or no arguments hides all the subwidgets
        """

        hide_all = len(args) == 0 or args[0] == "all"

        if hide_all or "inclusion" in args:
            self.inclusion.grid_forget()
        if hide_all or "field" in args:
            self.field.grid_forget()
        if hide_all or "operator" in args:
            self.operator.grid_forget()
        if hide_all or "value" in args:
            self.value.grid_forget()
        if hide_all or "value2" in args:
            self.value2.grid_forget()

    def set(self, *args):
        """Set the values of the widgets as a tuple.

        Parameters
        ----------
        args : [str] or str, str, str, str ,str
            either a single iterable of length 5 or 5 separate values...
        """
        if len(args) == 1:
            if len(args[0]) == 5:
                args = args[0]
            else:
                raise ValueError("set expects with a 5-long list or 5 separate values")
        elif len(args) != 5:
            raise ValueError("set expects with a 5-long list or 5 separate values")

        self.inclusion.set(args[0])
        self.field.set(args[1])
        self.operator.set(args[2])
        self.value.set(args[3])
        self.value2.set(args[4])

        self.callback("internal", "set")

    def show(self, *args):
        """Show the specified subwidgets.
        'all' or no arguments reverts to showing all"""

        show_all = len(args) == 0 or args[0] == "all"

        if show_all or "inclusion" in args:
            self.inclusion.grid(row=0, column=0, sticky=tk.EW)
        if show_all or "field" in args:
            self.field.grid(row=0, column=1, sticky=tk.EW)
        if show_all or "operator" in args:
            self.operator.grid(row=0, column=2, sticky=tk.EW)
        if show_all or "value" in args:
            self.value.grid(row=0, column=3, sticky=tk.EW)
        if show_all or "value2" in args:
            self.value2.grid(row=0, column=4, sticky=tk.EW)


class SearchCriteria(sw.ScrolledLabelFrame):
    """Class to editable list of criteria for e.g. searching."""

    def __init__(self, parent, *args, **kwargs):
        class_ = kwargs.pop("class_", "MSearchCriteria")
        super().__init__(parent, class_=class_)

        self.command = kwargs.pop("command", None)

        self.inclusiontext = kwargs.pop("inclusiontext", "")
        self.inclusionvalues = kwargs.pop("inclusionvalues", ("with", "without"))
        self.inclusionheight = kwargs.pop("inclusionheight", 7)
        self.inclusionwidth = kwargs.pop("inclusionwidth", 10)
        self.inclusionstate = kwargs.pop("inclusionstate", "readonly")

        self.fieldtext = kwargs.pop("fieldtext", "")
        self.fieldvalues = kwargs.pop("fieldvalues", tuple())
        self.fieldheight = kwargs.pop("fieldheight", 7)
        self.fieldwidth = kwargs.pop("fieldwidth", 10)
        self.fieldstate = kwargs.pop("fieldstate", "readonly")

        self.operatortext = kwargs.pop("operatortext", "")
        self.operatorvalues = kwargs.pop("operatorvalues", tuple())
        self.operatorheight = kwargs.pop("operatorheight", 7)
        self.operatorwidth = kwargs.pop("operatorwidth", 10)
        self.operatorstate = kwargs.pop("operatorstate", "readonly")

        self.two_values = kwargs.pop("two_values", ("between",))

        self._rows = []

        frame = self.frame = self.interior()

        # The button to add a new row
        self.add_button = ttk.Button(
            frame,
            text="+",
            width=2,
            command=self.add_row,
            takefocus=True,
        )

        # interior frame
        self.interior = ttk.Frame(frame)
        self.interior.grid(row=0, column=5, sticky=tk.NSEW)

        self.config(**kwargs)

        self.layout()

    def add_row(self, values=None):
        """Add a new row to the table.

        Parameters
        ----------
        values : [5]
            Values to set the criterion widget with.
        """
        # The button to remove a row...
        remove_button = ttk.Button(
            self.frame,
            text="-",
            width=2,
            command=lambda index=len(self._rows): self.remove_row(index),
            takefocus=True,
        )

        # and the criterion
        criterion = Criterion(
            self.frame,
            command=self.callback,
            inclusiontext=self.inclusiontext,
            inclusionvalues=self.inclusionvalues,
            inclusionheight=self.inclusionheight,
            inclusionwidth=self.inclusionwidth,
            inclusionstate=self.inclusionstate,
            fieldtext=self.fieldtext,
            fieldvalues=self.fieldvalues,
            fieldheight=self.fieldheight,
            fieldwidth=self.fieldwidth,
            fieldstate=self.fieldstate,
            operatortext=self.operatortext,
            operatorvalues=self.operatorvalues,
            operatorheight=self.operatorheight,
            operatorwidth=self.operatorwidth,
            operatorstate=self.operatorstate,
            two_values=self.two_values,
        )
        if values is not None:
            criterion.set(values)

        self._rows.append((remove_button, criterion))

        self.callback(criterion, "internal", "add row")

        self.layout()

        return criterion

    def callback(self, criterion, event, what):
        """Call the command for changes in the widget value."""
        if self.command is not None:
            self.command(self, criterion, event, what)

    def get(self):
        """Return the list of criteria."""
        result = []
        for _, criterion in self._rows:
            result.append(criterion.get())

        return result

    def layout(self):
        """Layout the criteria.

        Indent rows after a parenthesis.
        """

        frame = self.frame

        # Remove any weights on columns
        for i in range(frame.grid_size()[0]):
            frame.grid_columnconfigure(i, weight=0)

        # Unpack any widgets
        for slave in frame.grid_slaves():
            slave.grid_forget()

        level = 0
        max_level = 0
        for remove_button, criterion in self._rows:
            inclusion = criterion.get()[0]
            if "(" in inclusion:
                level += 1
                if level > max_level:
                    max_level = level
            elif ")" in inclusion:
                level -= 1

        level = 0
        row = 0
        for remove_button, criterion in self._rows:
            remove_button.grid(row=row, column=0, sticky=tk.EW)
            span = max_level - level + 1
            logger.debug(
                f"criterion.grid(row={row}, column={1 + level}, columnspan={span}..."
            )
            criterion.grid(row=row, column=1 + level, columnspan=span, sticky=tk.EW)
            row += 1

            inclusion = criterion.get()[0]
            if "(" in inclusion:
                level += 1
                if level > max_level:
                    max_level = level
            elif ")" in inclusion:
                level -= 1
            logger.debug(f"{inclusion=}, {level=}")

        self.add_button.grid(row=row, column=0, sticky=tk.EW)

        # Set up the last column as the one that expands
        frame.grid_columnconfigure(max_level + 1, weight=1)

        # and minumum widths to indent
        for i in range(1, max_level + 1):
            logger.debug(f"frame.grid_columnconfigure({i}, minsize=30)")
            frame.grid_columnconfigure(i, minsize=30)

    def remove_row(self, index):
        """Remove a row from the table.

        Parameters
        ----------
        index : int
            The row in the table to remove.
        """
        remove_button, criterion = self._rows[index]
        remove_button.destroy()
        criterion.destroy()
        del self._rows[index]

        self.callback(None, "internal", "remove row")

        # Set the index of each row...
        for i, row in enumerate(self._rows):
            row[0].configure(command=lambda index=i: self.remove_row(index))

        self.layout()

    def set(self, criteria):
        """Set the current value of the widget to the given criteria.

        Parameters
        ----------
        criteria : [tuple()]
            A list of the value for each criterion.
        """
        self.clear()

        for values in criteria:
            self.add_row(values=values)


if __name__ == "__main__":  # pragma: no cover

    def check(w):
        print(w.get())

    def set_values(w):
        w.set("with out", "author", "between", "a", "m")

    def callback(self, criterion, event, what):
        if what in ("add row", "field", "set"):
            inclusion, field, operator, value, value2 = criterion.get()
            w = criterion.operator
            if field == "author":
                w.configure(values=("like", "is"))
                w.set("like")
            elif field == "keyword":
                w.configure(values=("is", "like"))
                w.set("is")
            elif field == "date":
                w.configure(values=("is", "is before", "is after", "between"))
                w.set("is after")

    root = tk.Tk()
    root.title("Search Criteria")

    # criterion = Criterion(
    #     root,
    #     fieldvalues=("keyword", "author", "date"),
    #     operatorvalues=("=", "like"),
    #     two_values=("between",),
    # )
    # criterion.grid(row=0, column=0, columnspan=2, sticky=tk.EW)

    # w = ttk.Button(root, text="Check", command=lambda w=criterion: check(w))
    # w.grid(row=1, column=0)

    # w = ttk.Button(root, text="Set", command=lambda w=criterion: set_values(w))
    # w.grid(row=1, column=1)

    search = SearchCriteria(
        root,
        text="Find Flowcharts",
        labelanchor=tk.N,
        inclusiontext="That ",
        inclusionvalues=("contain", "do not contain"),
        fieldvalues=("keyword", "author", "date"),
        operatorvalues=("<", ">", "=", "like", "between"),
        two_values=("between",),
        command=callback,
    )
    search.grid(row=0, column=0, columnspan=2, sticky=tk.NSEW)

    root.grid_rowconfigure(0, weight=1)
    root.grid_columnconfigure(0, weight=1)

    root.mainloop()



# -*- coding: utf-8 -*-

"""Labeled combobox widget with units.

The goal of these widgets is twofold: to make it easier for developers
to implement dialogs with compound widgets, and to naturally
standardize the user interface presented to the user.
"""

import logging
from seamm_util import Q_, units_class, default_units
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk

logger = logging.getLogger(__name__)

options = {
    "unitcombobox": {
        "as_quantity": "as_quantity",
    },
    "units": {
        "cursor": "cursor",
        "exportselection": "exportselection",
        "unitsheight": "height",
        "unitsjustify": "justify",
        "postcommand": "postcommand",
        "style": "style",
        "unitstakefocus": "takefocus",
        "variable": "textvariable",
        "unitsvalidate": "validate",
        "unitsvalidatecommand": "validatecommand",
        "unitswidth": "width",
        "unitsxscrollcommand": "xscrollcommand",
        "unitsstate": "state",
    },
}


class UnitCombobox(sw.LabeledCombobox):
    def __init__(self, parent, *args, **kwargs):
        """Initialize the instance"""
        # Pull out the specific unitcombobox options
        self.as_quantity = kwargs.pop("as_quantity", False)

        myoptions = {
            "height": 7,
            "width": 10,
            "state": "readonly",
        }
        for option, myoption in options["units"].items():
            if option in kwargs:
                myoptions[myoption] = kwargs.pop(option)

        # Create our parent
        class_ = kwargs.pop("class_", "MUnitCombobox")
        super().__init__(parent, class_=class_, *args, **kwargs)

        interior = self.interior

        # And put our widget in
        self.units = ttk.Combobox(interior, **myoptions)
        self.units.grid(row=0, column=0, sticky=tk.EW)

        # interior frame
        self.interior = ttk.Frame(interior)
        self.interior.grid(row=0, column=1, sticky=tk.NSEW)
        interior.columnconfigure(0, weight=1)

    @property
    def value(self):
        return self.get()

    @value.setter
    def value(self, value):
        self.set(value)

    def show(self, *args):
        """Show only the specified subwidgets.
        'all' or no arguments reverts to showing all"""

        super().show(*args)

        show_all = len(args) == 0 or args[0] == "all"

        if show_all or "units" in args:
            self.units.grid_remove()
            self.units.grid()
        else:
            self.units.grid_remove()

    def set(self, value, unit_string=None):
        """Set the the value and units"""

        if value is None:
            return

        # the value may have units or be a plain value
        if isinstance(value, units_class):
            self.combobox.set(value.magnitude)

            dimensionality = value.dimensionality
            current_units = self.units.cget("values")
            if len(current_units) > 0:
                for unit in current_units:
                    if unit != "":
                        if Q_(unit).dimensionality != dimensionality:
                            self.units.configure(values=[])
                            current_units = []
                            break

            if len(current_units) == 0:
                self.set_units([*default_units(str(dimensionality)), ""])
                self.units.set("{0.units:~}".format(value).replace(" ", ""))
        elif unit_string is not None:
            self.combobox.set(value)

            dimensionality = Q_(unit_string).dimensionality
            current_units = self.units.cget("values")
            if len(current_units) > 0:
                for unit in current_units:
                    if unit != "":
                        if Q_(unit).dimensionality != dimensionality:
                            self.units.configure(values=[])
                            current_units = []
                            break

            if len(current_units) == 0:
                self.set_units([*default_units(str(dimensionality)), ""])
                self.units.set(unit_string)
        else:
            self.combobox.set(value)
            self.set_units("all")
            self.units.set("")

    def get(self):
        """return the current value with units"""
        value = self.combobox.get()
        if value in self.combobox.cget("values"):
            return value
        else:
            unit = self.units.get()
            if unit == "":
                return value
            elif self.as_quantity:
                try:
                    magnitude = float(value)
                    return Q_(magnitude, unit)
                except Exception:
                    return (value, unit)
            else:
                return (value, unit)

    def set_units(self, values=None):
        if values is None:
            dimensionality = str(self.get().dimensionality)
            self.units.config(values=[*default_units(dimensionality), ""])
        elif values == "all":
            self.units.config(values=[*default_units("all"), ""])
        else:
            self.units.config(values=values)

    def config(self, **kwargs):
        """Set the configuration of the megawidget"""

        unitcombobox = options["unitcombobox"]
        units = options["units"]

        # cannot modify kwargs while iterating over it...
        keys = [*kwargs.keys()]
        for k in keys:
            if k in unitcombobox and unitcombobox[k] in self.__dict__:
                v = kwargs.pop(k)
                self.__dict__[unitcombobox[k]] = v
            elif k in units:
                v = kwargs.pop(k)
                self.units.config(**{units[k]: v})

        # having removed our options, pass rest to parent
        super().config(**kwargs)

    def configure(self, **kwargs):
        return self.config(**kwargs)

    def state(self, stateSpec=None):
        """Set the state of the widget"""
        result = super().state(stateSpec)
        tmp = self.units.state(stateSpec)
        return result + tmp



# -*- coding: utf-8 -*-

"""A Tk scrolled frame containing and handling a table of properties.

This megawidget displays a list of properties that the user can add to and also
delete.
"""

from itertools import takewhile
import logging
import Pmw
import seamm_widgets as sw
import tkinter as tk
import tkinter.ttk as ttk

module_logger = logging.getLogger(__name__)


def lcp(*s):
    """Longest common prefix of strings"""
    return "".join(a for a, b in takewhile(lambda x: x[0] == x[1], zip(min(s), max(s))))


class PropertyTable(sw.ScrolledFrame):
    """A widget to handle manual input of property names"""

    def __init__(
        self,
        master,
        metadata=None,
        properties="",
        logger=module_logger,
        labeltext="",
        **kwargs
    ):
        """ """
        self.logger = logger
        self._properties = []
        self._max_width = 0
        self._working_properties = None
        self._add_widget = None
        self._dialog = None
        self._set_property_cb = None
        self._labeltext = labeltext
        self.popup_menu = None

        class_ = kwargs.pop("class_", "MProperties")

        s = ttk.Style()
        s.configure("Red.TEntry", foreground="red")

        super().__init__(master, class_=class_, inner_frame=ttk.Frame, **kwargs)

        # After everything is set up can put in the properties and metadata
        self.metadata = metadata
        self.properties = properties

    @property
    def properties(self):
        """The current properties, reflecting changes in the widgets"""
        # result = []
        # for d in self._working_properties:
        #     result.append((d['property'], d['accuracy']))
        # return result
        return self.get()

    @properties.setter
    def properties(self, value):
        self.logger.debug("Properties::properties: " + str(value))
        self.clear()
        self._working_properties = []
        self._properties = value
        if value is not None:
            for item in value:
                _property, accuracy = item
                self._working_properties.append(
                    {"property": _property, "accuracy": accuracy}
                )
        self.logger.debug("  properties.setter call layout_properties")
        self.layout_properties()

    def set(self, value):
        """Set the properties in the widget."""
        self.properties = value

    @property
    def metadata(self):
        """The current metadata"""
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        self._metadata = value
        self._max_width = 0
        if value is not None:
            for _property in self._metadata.keys():
                width = len(_property)
                if width > self._max_width:
                    self._max_width = width
            self._max_width += 3  # a little padding...
            self.properties = self._properties

    def clear(self):
        """Remove the widgets."""
        if self._working_properties is not None:
            for d in self._working_properties:
                if "widgets" in d:
                    for widget in d["widgets"].values():
                        widget.destroy()
                    del d["widgets"]

    def reset(self):
        """Remove any changes made in the dialog."""
        self.properties = self._properties

    def layout_properties(self):
        """Layout the table of additional properties and any arguments they
        need"""
        self.logger.debug("Properties::layout_properties")

        frame = self.innerframe

        # Unpack any widgets
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # Callbacks
        if self._set_property_cb is None:
            self._set_property_cb = frame.register(self.set_property_cb)

        row = -1
        for d in self._working_properties:
            self.logger.debug(d)
            row += 1
            _property = d["property"]
            if "widgets" not in d:
                widgets = d["widgets"] = {}
            else:
                widgets = d["widgets"]

            if "remove" not in widgets:
                # The button to remove a row...
                widgets["remove"] = ttk.Button(
                    frame,
                    text="-",
                    width=2,
                    command=lambda row=row: self.remove_property(row),
                    takefocus=True,
                )

            if "name" not in widgets:
                # the name of the property
                widgets["name"] = ttk.Label(
                    frame, width=self._max_width, text=_property
                )
                widgets["plusminus"] = ttk.Label(frame, text="±")
                # and desired accuracy
                widgets["accuracy"] = sw.UnitEntry(frame)
                if "accuracy" in d:
                    accuracy, units = d["accuracy"]
                else:
                    units = self._metadata[_property]["units"]
                    if "accuracy" in self._metadata[_property]:
                        accuracy = self._metadata[_property]["accuracy"]
                    else:
                        accuracy = "0.1%"
                widgets["accuracy"].set(accuracy, units)

            self.logger.debug("  widgets: " + str(widgets))
            widgets["remove"].grid(row=row, column=0, sticky=tk.W)
            col = 1
            widgets["name"].grid(row=row, column=col, stick=tk.EW)
            col += 1
            widgets["plusminus"].grid(row=row, column=col, stick=tk.EW)
            col += 1
            widgets["accuracy"].grid(row=row, column=col, stick=tk.EW)
            col += 1

        # The button to add a row...
        row += 1
        if self._add_widget is None:
            self._add_widget = ttk.Button(
                frame,
                text="+",
                width=5,
                command=self.add_property,
                takefocus=True,
            )
            self._add_widget.focus_set()
        self._add_widget.lift()
        self._add_widget.grid(row=row, column=0, columnspan=3, sticky=tk.W)

        frame.grid_columnconfigure(2, weight=1)

    def get(self):
        """Get the values of the properties from the widgets"""
        self.logger.debug("Properties::get_properties")

        properties = []
        for d in self._working_properties:
            widgets = d["widgets"]
            _property = widgets["name"].cget("text")
            accuracy = widgets["accuracy"].get()

            properties.append((_property, accuracy))

        return properties

    def set_property_cb(self, _property):
        self.logger.debug("Properties::set_property_cb")
        self.logger.debug(_property)

        self._working_properties.append({"property": _property})
        self.layout_properties()

    def add_property(self, _property=""):
        """Add a property to the input"""
        # Post a menu with the choices
        popup_menu = tk.Menu(self.innerframe, tearoff=0)
        current = []
        for d, _ in self.properties:
            current.append(d)

        for _property in self._metadata:
            if _property not in current:
                description = self._metadata[_property]["description"]
                popup_menu.add_command(
                    label="{}: {}".format(_property, description),
                    command=(self._set_property_cb, _property),
                )
        x, y = self.winfo_pointerxy()
        popup_menu.tk_popup(x, y, 0)

    def remove_property(self, row=None):
        """Remove a property from dd to input"""
        self.logger.debug("remove row {}".format(row))
        for widget in self._working_properties[row]["widgets"].values():
            widget.destroy()
        del self._working_properties[row]
        self.layout_properties()


if __name__ == "__main__":  # pragma: no cover
    import sys

    metadata = {
        "T": {
            "calculation": [
                "nve",
                "nvt",
                "npt",
            ],
            "description": "temperature",
            "dimensionality": "scalar",
            "type": "float",
            "units": "K",
            "accuracy": "2.0",
        },
        "P": {
            "calculation": [
                "nve",
                "nvt",
                "npt",
            ],
            "description": "pressure",
            "dimensionality": "scalar",
            "type": "float",
            "units": "atm",
        },
        "density": {
            "calculation": [
                "nve",
                "nvt",
                "npt",
            ],
            "description": "pressure",
            "dimensionality": "scalar",
            "type": "float",
            "units": "g/ml",
        },
        "a": {
            "calculation": [
                "nve",
                "nvt",
                "npt",
            ],
            "description": "cell parameter 'a'",
            "dimensionality": "scalar",
            "type": "float",
            "units": "Å",
        },
        "b": {
            "calculation": [
                "nve",
                "nvt",
                "npt",
            ],
            "description": "cell parameter 'b'",
            "dimensionality": "scalar",
            "type": "float",
            "units": "Å",
        },
        "c": {
            "calculation": [
                "nve",
                "nvt",
                "npt",
            ],
            "description": "cell parameter 'c'",
            "dimensionality": "scalar",
            "type": "float",
            "units": "Å",
        },
        "Etot": {
            "calculation": [
                "nve",
                "nvt",
                "npt",
            ],
            "description": "total energy",
            "dimensionality": "scalar",
            "type": "float",
            "units": "kcal/mol",
        },
        "Eke": {
            "calculation": [
                "nve",
                "nvt",
                "npt",
            ],
            "description": "kinetic energy",
            "dimensionality": "scalar",
            "type": "float",
            "units": "kcal/mol",
        },
        "Epe": {
            "calculation": [
                "nve",
                "nvt",
                "npt",
            ],
            "description": "potential energy",
            "dimensionality": "scalar",
            "type": "float",
            "units": "kcal/mol",
        },
        "Epair": {
            "calculation": [
                "nve",
                "nvt",
                "npt",
            ],
            "description": "nonbonded (vdW & electrostatic) energy",
            "dimensionality": "scalar",
            "type": "float",
            "units": "kcal/mol",
        },
    }

    # Helper to print the current properties
    def print_properties(*args):
        print(w.get())

    def print_focus(*args):
        print(root.focus_get())

    def handle_dialog(result):
        dialog.deactivate(result)
        if result == "OK":
            properties = w.get()
            print("OK")
            print(properties)
            w.properties = properties
        else:
            w.reset()
            print("Cancel")
            print(w.properties)

    ##################################################
    # Initialize Tk
    ##################################################
    if sys.platform.startswith("darwin"):
        CmdKey = "Command-"
    else:
        CmdKey = "Control-"

    #    logging.basicConfig(level=10)
    #    module_logger.critical('Turned on debugging!')
    #    module_logger.setLevel('DEBUG')
    #    module_logger.setLevel(10)
    #    print(module_logger.getEffectiveLevel())
    #    module_logger.debug('Turned on debugging!')

    root = tk.Tk()
    Pmw.initialise(root)

    dialog = Pmw.Dialog(
        root,
        buttons=("OK", "Cancel"),
        defaultbutton=None,
        master=root,
        title="Add property",
        command=handle_dialog,
    )
    dialog.geometry("500x400")

    properties = [("P", (0.003, "atm"))]

    w = PropertyTable(dialog.interior(), metadata)
    w.properties = properties
    w.pack(expand="yes", fill="both")

    root.bind_all("<" + CmdKey + "P>", print_properties)
    root.bind_all("<" + CmdKey + "p>", print_properties)
    root.bind_all("<" + CmdKey + "f>", print_focus)

    exit_button = ttk.Button(root, text="Exit", command=exit)
    post_button = ttk.Button(root, text="Edit Properties", command=dialog.activate)
    post_button.grid(column=0, row=0)
    exit_button.grid(column=1, row=0)

    dialog.activate(geometry="centerscreenfirst")

    # enter the event loop
    root.mainloop()
